/***********************************************************************[pfsolve.h]
Copyright(c) 2020, Muhammad Osama - Anton Wijs,
Technische Universiteit Eindhoven (TU/e).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************************/

#ifndef __SOLVE_
#define __SOLVE_

#include "pfvec.h"
#include "pfargs.h"
#include "pfheap.h"
#include "pfqueue.h"
#include "pfdefs.h"

//====================================================//
//       Solver Data structures primitives            //
//====================================================//

/*****************************************************/
/*  Name:     SOL                                    */
/*  Usage:    store assignment info.                 */
/*  Scope:    host                                   */
/*  Memory:   system memory                          */
/*  Dependency:  none                                */
/*****************************************************/
class SOL
{
	LIT_ST* _values;
	int* _level;
	uint32 numVars;
public:
	inline SOL() { _values = NULL, _level = NULL, numVars = 0; }
	inline ~SOL() { _values = NULL, _level = NULL, numVars = 0; }
	inline void allocMem(addr_t* _mem, const uint32& sz) {
		numVars = sz;
		assert(numVars > 0);
		_values = (LIT_ST*)(*_mem);
		*_mem += numVars * sizeof(LIT_ST);
		_level = (int*)(*_mem);
		*_mem += numVars * sizeof(int);
	}
	inline operator LIT_ST* (void) { return  _values; }
	inline uint32 size() const { return numVars; }
	inline void init(const uint32& idx) { _values[idx] = UNDEFINED, _level[idx] = UNDEFINED; }
	inline void init_value(const uint32& idx) { _values[idx] = UNDEFINED; }
	inline void init_level(const uint32& idx) { _level[idx] = UNDEFINED; }
	inline void set_value(const uint32& idx, const LIT_ST& asg) { _values[idx] = asg; }
	inline LIT_ST value(const uint32& idx) { return _values[idx]; }
	inline int level(const uint32& idx) { return _level[idx]; }
	inline void set_level(const uint32& idx, const int& dl) { _level[idx] = dl; }
	inline void set_var(const uint32& idx, const LIT_ST& asg, const int& dl) {
		_values[idx] = asg;
		_level[idx] = dl;
	}
	inline void print() {
		printf("c | Assigned vars: \n");
		for (uint32 i = 0; i < numVars; i++) {
			if (_values[i] != UNDEFINED)
				printf("c | var(%d@%d) = %d\n", i + 1, _level[i], (int)_values[i]);
		}
		printf("\n");
	}
};
/*****************************************************/
/*  Name:     BCLAUSE                                */
/*  Usage:    abstract clause for solving on host    */
/*  Dependency:  none                                */
/*****************************************************/
class BCLAUSE {
	uint32* _lits;
	CL_LEN _sz;
	int8_t _st;
public:
	inline BCLAUSE() { _lits = NULL, _sz = UNKNOWN, _st = UNKNOWN; }
	inline explicit BCLAUSE(const CL_LEN& newSz) {
		_lits = NULL, _st = UNKNOWN;
		_sz = newSz;
		_lits = new uint32[_sz];
	}
	inline virtual ~BCLAUSE() { clear(true); }
	inline uint32& operator [] (LIT_POS i) { assert(i < _sz); return _lits[i]; }
	inline uint32  operator [] (LIT_POS i) const { assert(i < _sz); return _lits[i]; }
	inline operator uint32* (void) { assert(_sz != 0); return _lits; }
	inline uint32* d_ptr() { return _lits; }
	inline CL_LEN size() const { return _sz; }
	inline void pop() { _sz--; }
	inline void shrink(const CL_LEN& n) { _sz -= n; }
	inline void resize(const CL_LEN& newSz) { _sz = newSz; }
	inline void set_status(const int8_t& status) { assert(status <= ST_MASK); _st = (_st & ST_RST) | status; }
	inline void markDeleted() { assert(DELETED == ST_MASK); _st |= DELETED; }
	inline int8_t status() const { return (_st & ST_MASK); }
	inline void initGarbage() { _st &= GAR_RST; }
	inline void markGarbage() { _st |= GAR_MASK; }
	inline bool garbage() const { return (_st & GAR_MASK); }
	inline void initReason() { _st &= IMP_RST; }
	inline void markReason(void) { _st |= IMP_MASK; }
	inline bool reason() const { return (_st & IMP_MASK); }
	inline void initOrgBin(void) { _st &= BIN_RST; }
	inline void markOrgBin(void) { _st |= BIN_MASK; } // set as original binary
	inline bool orgBin(void) const { return (_st & BIN_MASK); }
	inline void melt(void) { _st |= DEL_MASK; } // set for deletion
	inline void freeze() { _st &= DEL_RST; } // prevent from deletion
	inline bool molten() const { return (_st & DEL_MASK); } // is removable
	inline bool isWatchedVar(const uint32& v_idx) const { assert(_sz > 1); return (V2X(*_lits) == v_idx || V2X(*(_lits + 1)) == v_idx); }
	inline void copyLitsFrom(uVec1D& src) {
		assert(_sz <= src.size());
		uint32* d = _lits, *end = d + _sz, *s = src;
		while (d != end) *d++ = *s++;
		assert(int(d - _lits) == _sz);
	}
	inline bool satisfied(const LIT_ST* values) const {
		if (status() == DELETED) return true;
		uint32* h = _lits;
		uint32* e = h + _sz;
		while (h != e) {
			assert(*h > 0);
			if (values[V2X(*h)] == !ISNEG(*h)) return true;
			h++;
		}
		return false;
	}
	inline void swap_ws() { // swap watched literals
		uint32 lit0 = *_lits;
		*_lits = _lits[1];
		_lits[1] = lit0;
	}
	inline void swap(const LIT_POS& p1, const LIT_POS& p2) { // swap two literals
		uint32 tmp = _lits[p1];
		_lits[p1] = _lits[p2];
		_lits[p2] = tmp;
	}
	inline void clear(bool _free = false) {
		if (_free && _lits != NULL) { delete _lits; _lits = NULL; }
		_sz = UNKNOWN; _st = UNKNOWN;
	}
	inline void print() const {
		printf("(");
		for (LIT_POS l = 0; l < _sz; l++) {
			int lit = int(ABS(_lits[l]));
			lit = (ISNEG(_lits[l])) ? -lit : lit;
			printf("%4d ", lit);
		}
		if (this->status() == ORIGINAL) printf(") O, r=%d\n", reason());
		else if (this->status() == LEARNT) printf(") A, r=%d\n", reason());
		else if (this->status() == DELETED) printf(") X, r=%d\n", reason());
		else printf(") U, r=%d\n", reason());
	}
};
typedef BCLAUSE* B_REF; // ref to BCLAUSE
typedef Vec<B_REF> BCNF;
/*****************************************************/
/*  Name:     LCLAUSE                                */
/*  Usage:    Extended BCLAUSE with clause heuristics*/
/*  Dependency:  BCLAUSE                             */
/*****************************************************/
class LCLAUSE: public BCLAUSE {
	float _act;
	uint32 _lbd;
public:
	inline LCLAUSE() : BCLAUSE(), _lbd(UNKNOWN), _act(0.0) {}
	inline explicit LCLAUSE(const CL_LEN& sz) : BCLAUSE(sz), _lbd(UNKNOWN), _act(0.0) {}
	inline ~LCLAUSE() { _act = 0.0; _lbd = UNKNOWN; }
	inline void set_LBD(const uint32& lbd) { _lbd = lbd; }
	inline uint32 LBD() const  { return _lbd; }
	inline float activity() const { return _act; }
	inline void inc_act(const float& act) { _act += act; }
	inline void sca_act(const float& act) {	_act *= act; }
	inline void set_act(const float& act) { _act = act; }
	inline void print() {
		printf("(");
		for (LIT_POS l = 0; l < size(); l++) {
			int lit = int(ABS((*this)[l]));
			lit = (ISNEG((*this)[l])) ? -lit : lit;
			printf("%4d ", lit);
		}
		if (this->status() == ORIGINAL) printf(") O:%d, g=%d, a=%.2f\n", reason(), LBD(), activity());
		else if (this->status() == LEARNT) printf(") A:%d, g=%d, a=%.2f\n", reason(), LBD(), activity());
		else if (this->status() == DELETED) printf(") X:%d, g=%d, a=%.2f\n", reason(), LBD(), activity());
		else printf(") U:%d, g=%d, a=%.2f\n", reason(), LBD(), activity());
	}
};
typedef LCLAUSE* C_REF; // ref to LCLAUSE
typedef Vec<C_REF> LCNF;
/*****************************************************/
/*  Name:     SCLAUSE                                */
/*  Usage:    Extended BCLAUSE with signature        */
/*  Dependency:  BCLAUSE                             */
/*****************************************************/
class SCLAUSE : public BCLAUSE {
	uint32 _sig;
public:
	inline SCLAUSE() : BCLAUSE(), _sig(UNKNOWN) {}
	inline explicit SCLAUSE(const CL_LEN& sz) : BCLAUSE(sz), _sig(UNKNOWN) {}
	inline ~SCLAUSE() { _sig = UNKNOWN; }
	inline uint32 lit(const LIT_POS& idx) { assert(idx < size()); return (*this)[idx]; }
	inline void set_sig(const uint32& sig) { _sig = sig; }
	inline uint32 sig() { return _sig; }
	inline LIT_POS hasZero() {
		for (LIT_POS l = 0; l < size(); l++) {
			if (this->lit(l) == 0) return l;
		}
		return -1;
	}
	inline void calcSig(const uint32& init_sig = 0) {
		_sig = init_sig;
		for (LIT_POS l = 0; l < this->size(); l++)
			_sig |= MAPHASH((*this)[l]);
	}
	inline void print() {
		printf("(");
		for (LIT_POS l = 0; l < size(); l++) {
			int lit = int(ABS((*this)[l]));
			lit = (ISNEG((*this)[l])) ? -lit : lit;
			printf("%4d ", lit);
		}
		if (this->status() == ORIGINAL) printf(") O, s=0x%X\n", sig());
		else if (this->status() == LEARNT) printf(") A, s=0x%X\n", sig());
		else if (this->status() == DELETED) printf(") X, s=0x%X\n", sig());
		else printf(") U, s=0x%X\n", sig());
	}
	inline void copyFrom(BCLAUSE* src) {
		assert(size() <= src->size());
		uint32* d = d_ptr(), *end = d + size(), *s = src->d_ptr();
		while (d != end) *d++ = *s++;
		assert(int(d - d_ptr()) == size());
	}
	inline void copyTo(BCLAUSE* dest) {
		assert(dest->size() <= size());
		uint32* d = dest->d_ptr(), *end = d + size(), *s = d_ptr();
		while (d != end) *d++ = *s++;
		assert(int(d - dest->d_ptr()) == dest->size());
	}
	// binary search
	inline bool has(const uint32& lit) { 
		if (size() == 2) {
			if (this->lit(0) == lit || this->lit(1) == lit) return true;
			else return false;
		}
		else {
			assert(this->isSorted());
			LIT_POS low = 0, high = size() - 1, mid;
			uint32 first = this->lit(low), last = this->lit(high);
			while (first <= lit && last >= lit) {
				mid = (low + high) >> 1;
				uint32 m = this->lit(mid);
				if (m < lit) first = this->lit(low = mid + 1);
				else if (m > lit) last = this->lit(high = mid - 1);
				else return true; // found
			}
			if (this->lit(low) == lit) return true; // found
			else return false; // Not found
		}
	}
	inline bool isSorted() {
		for (LIT_POS i = 0; i < size(); i++) {
			if (i > 0 && this->lit(i) < this->lit(i - 1)) return false;
		}
		return true;
	}
};
typedef SCLAUSE* S_REF; // ref to SCLAUSE
typedef Vec<S_REF> SCNF;
typedef Vec<S_REF> OL;
typedef Vec<OL, int64> OT;
/*****************************************************/
/*  Name:     Watch table                            */
/*  Usage:    stores pointers to clauses per literal */
/*            with builtin garbage collector         */
/*  Dependency:  Watch                               */
/*****************************************************/
struct WATCH {
	G_REF c_ref;
	uint32  imp;
	WATCH() { c_ref = NULL; imp = 0; }
	WATCH(G_REF ref, uint32 lit) { c_ref = ref; imp = lit; }
	~WATCH() { c_ref = NULL; imp = 0; }
};
typedef Vec<WATCH> WL;
class WT {
	Vec<WL, int64> wt;
	Vec<uint32> garbage;
	Vec<bool> collected;
	uint32 nVars;
public:
	inline WT() { nVars = 0; }
	inline WL& operator[](const uint32& lit) { assert(lit > 1); return wt[lit]; }
	inline const WL& operator[](const uint32& lit) const { assert(lit > 1); return wt[lit]; }
	inline void allocMem(const uint32& sz) { assert(sz); nVars = sz; wt.resize((int64)V2D(nVars + 1LL)); collected.resize((int64)V2D(nVars + 1LL), 0); }
	inline int64 size() const { return wt.size(); }
	inline bool empty() const  { return wt.size() == 0; }
	inline void destroy(const uint32& lit) { wt[lit].clear(true); }
	inline WL& getClean(const uint32& lit) { assert(lit > 1); if (collected[lit]) recycle(lit); return wt[lit]; }
	inline void collect(const uint32& lit) { assert(lit > 1); if (!collected[lit]) { collected[lit] = 1; garbage.push(lit); } }
	inline void recycle(const uint32& lit) {
		WL& ws = wt[lit];
		if (ws.empty()) { ws.clear(true); collected[lit] = 0; return; }
		WATCH* w_i, *w_j, *end = ws + ws.size();
		for (w_i = ws, w_j = ws; w_i != end; w_i++)
			if (!(((B_REF)w_i->c_ref)->garbage())) *w_j++ = *w_i;
		ws.shrink(int(w_i - w_j));
		if (ws.empty()) ws.clear(true);
		collected[lit] = 0;
	}
	inline void recycle() {
		for (int i = 0; i < garbage.size(); i++) if (collected[garbage[i]]) recycle(garbage[i]);
		garbage.clear();
	}
	inline void remWatch(const uint32& lit, const G_REF gref)
	{
		WL& ws = wt[lit];
		if (ws.size() == 0) return;
		if (ws.size() > 1) {
			int c_idx = 0;
			while (ws[c_idx].c_ref != gref) c_idx++;
			assert(c_idx < ws.size());
			while (c_idx < ws.size() - 1) { ws[c_idx] = ws[c_idx + 1]; c_idx++; }
		}
		ws.pop();
	}
	inline void print(const uint32& lit) {
		WL& ws = wt[lit];
		for (int i = 0; i < ws.size(); i++) {
			printf("c | ");
			if (((B_REF)ws[i].c_ref)->status() == LEARNT) {
				C_REF cl = (C_REF)ws[i].c_ref;
				cl->print();
			}
			else
				((B_REF)ws[i].c_ref)->print();
		}
	}
	inline void print(void) {
		printf("c | Negative occurs:\n");
		for (uint32 v = 0; v < nVars; v++) {
			uint32 p = V2D(v + 1);
			if (wt[p].size() != 0) printf("c | Var(%d):\n", -int(v + 1)), print(p);
		}
		printf("c | Positive occurs:\n");
		for (uint32 v = 0; v < nVars; v++) {
			uint32 n = NEG(V2D(v + 1));
			if (wt[n].size() != 0) printf("c | Var(%d):\n", v + 1), print(n);
		}
	}
	inline void clear(bool _free = true) {
		wt.clear(_free);
		collected.clear(_free);
		garbage.clear(_free);
		nVars = 0;
	}
};

/*****************************************************/
/*  Name:     CNF Garbage collector                  */
/*  Usage:    manages deleted clauses                */
/*  Dependency:  all clause types                    */
/*****************************************************/
class GC {
	Vec<G_REF> garbage;
public:
	inline GC() {}
	inline void init(const int& cap = KBYTE) { garbage.reserve(cap, 0); }
	inline void clear(bool _free = false) { garbage.clear(_free); }
	inline int size() const { return garbage.size(); }
	inline bool empty() const { return garbage.size() == 0; }
	inline void collect(G_REF c) { garbage.push(c); }
	inline void recycle() {
		for (int i = 0; i < garbage.size(); i++) {
			assert(garbage[i] != NULL);
			if (((B_REF)garbage[i])->status() != LEARNT) delete (B_REF)garbage[i];
			else delete (C_REF)garbage[i];
			garbage[i] = NULL;
		}
		// start clean
		garbage.clear(true);
		init(); 
	}
	inline void print() {
		if (empty()) { printf("c | Garbage is empty\n"); return; }
		printf("c |\t\tgarbage (size = %d):\n", size());
		printf("c | %11s\t %10s\n", "address", "clause");
		for (int i = 0; i < size(); i++) {
			printf("c | %11llx->", (int64)garbage[i]);
			if (((B_REF)garbage[i])->status() != LEARNT) ((B_REF)garbage[i])->print();
			else ((C_REF)garbage[i])->print();
		}
		printf("c |--------------------------------------------------------------------------------------|\n");
	}
};
/*****************************************************/
/*  Name:     Containers for simp., search and learn */
/*  Dependency:  none                                */
/*****************************************************/
struct PV {
	uint32* PVs;
	bool* melted;
	int nPVs, mu_inc;
};
/*****************************************************/
/*  Name:     SP                                     */
/*  Usage:    Information of search space            */
/*  Scope:    host only                              */
/*  Memory:   system memory                          */
/*  Dependency: none                                 */
/*****************************************************/
struct SP {
	uVec1D assigns;
	bool* lock, * seen, * frozen, * pol;
	int bt_level, cbt_level;
	int assigns_head, assigns_offset;
	uint32 learnt_lbd;
	bool max1Found;
	void reset_assigns() {
		assigns.clear();
		assigns_head = assigns_offset = UNKNOWN;
	}
	void reset_level() {
		bt_level = ROOT_LEVEL;
		cbt_level = UNDEFINED;
		max1Found = false;
	}
};
/*****************************************************/
/*  Name:     STATS, LEARN, CL_PARAM                 */
/*  Usage:    Statistics & heuristics                */
/*  Scope:    host only                              */
/*  Memory:   system memory                          */
/*  Dependency: none                                 */
/*****************************************************/
struct STATS {
	int64 n_units;
	int64 n_props;
	int64 n_fuds;
	int64 n_pds;
	int64 tot_lits, max_lits;
	int pdm_calls;
	int nRestartStops;
	int ncbt, cbt;
	void reset() {
		pdm_calls = UNKNOWN;
		n_pds = UNKNOWN;
		n_props = UNKNOWN;
		n_units = UNKNOWN;
		n_fuds = UNKNOWN;
		max_lits = UNKNOWN;
		tot_lits = UNKNOWN;
		nRestartStops = UNKNOWN;
		ncbt = cbt = UNKNOWN;
	}
};
struct CL_PARAM {
	float cl_inc, cl_decay;
	void init() {
		cl_inc = 1.0;
		cl_decay = float(0.999);
	}
};
struct LEARN {
	int64 simp_props;
	int64 max_learnt_cls;
	int64 nClsReduce;
	int adjust_cnt;
	double size_factor_cls;
	double adjust_start_conf, adjust_conf;
	double size_inc, adjust_inc;
	void init(void) {
		size_inc = 1.1;
		adjust_inc = 1.5;
		adjust_start_conf = 100;
		size_factor_cls = 1.0 / 3.1;
		simp_props = UNKNOWN;
		adjust_conf = adjust_start_conf;
		adjust_cnt = (uint32)adjust_conf;
	}
};
/*****************************************************/
/*  Name:     ParaFROST                              */
/*  Usage:    global handler for solver/simplifier   */
/*  Scope:    host only                              */
/*  Memory:   system memory                          */
/*****************************************************/
class ParaFROST {
protected:
	string path;
	TIMER* timer;
	Vec<OCCUR> occurs; Vec<SCORE> scores;
	WT wt; BCNF orgs, bins; LCNF learnts; GC gcr;
	VAR_HEAP* var_heap; PV* pv; SP* sp; SOL* sol; G_REF* source; 
	int* board;	uint32* tmp_stack, *simpLearnt;
	uVec1D learnt_cl, learntLits; Vec1D assigns_len;
	LEARN lrn; CL_PARAM cl_params; STATS stats;
	addr_t sysMem; size_t sysMem_sz;
	int64 sysMemAvail, sysMemCons;
	int64 maxConflicts, nConflicts, pdm_div;
	int starts, restarts, R, ref_vars, reductions, marker;
	BQUEUE<uint32> lbdQ, trailQ;
	double lbdSum; 
	std::ofstream proofFile;
	bool intr, units_f;
public:
	ParaFROST(const string&);
	~ParaFROST(void);
	//============== inline methods ===============
	inline void interrupt(void) { intr = true; }
	inline bool interrupted(void) { return intr; }
	inline void write_proof(const Byte& byte) { proofFile << byte; }
	inline double drand(void) const { return ((double)rand() / (double)RAND_MAX); };
	inline void incDL(void) { assigns_len.push(sp->assigns.size()); }
	inline int DL(void) const { return assigns_len.size(); }
	inline void clDecayAct() { cl_params.cl_inc *= (1 / cl_params.cl_decay); }
	inline int64 cnfSize() { return (int64)nBins() + nClauses() + learnts.size(); }
	inline void enqueue(const uint32& assign, const int& pLevel = ROOT_LEVEL, const G_REF src = NULL)
	{
		assert(assign > 0);
		register uint32 assign_idx = V2X(assign);
		source[assign_idx] = src;
		sp->lock[assign_idx] = true;
		sp->assigns.push(assign);
		sol->set_var(assign_idx, !ISNEG(assign), pLevel);
		if (verbose >= 3) printf("c | New %s: %d@%d.\n", src == NULL ? "decision" : "imp", (ISNEG(assign)) ? -int(ABS(assign)) : ABS(assign), pLevel);
	}
	inline void collectClause(B_REF& c, const bool& gc = true) {
		if (gc) { assert(!c->garbage()); c->markGarbage(); gcr.collect(c); }
		else { assert(c != NULL); delete c; c = NULL; }
	}
	inline void collectClause(C_REF& c, const bool& gc = true) {
		if (gc) { assert(!c->garbage()); c->markGarbage(); gcr.collect(c); }
		else { assert(c != NULL); delete c; c = NULL; }
	}
	inline uint32 calcLBD(uVec1D& c) {
		marker++;
		register uint32 lbd = 0;
		for (LIT_POS i = 0; i < c.size(); i++) {
			int litLevel = sol->level(V2X(c[i]));
			if (board[litLevel] != marker) { board[litLevel] = marker; lbd++; }
		}
		return lbd;
	}
	inline uint32 calcLBD(C_REF& c) {
		marker++;
		register uint32 lbd = 0;
		for (LIT_POS i = 0; i < c->size(); i++) {
			int litLevel = sol->level(V2X((*c)[i]));
			if (board[litLevel] != marker) { board[litLevel] = marker; lbd++; }
		}
		return lbd;
	}
	inline void PDM_fuse() {
		if ((!pdm_freq && SH == 2 && starts % ((maxConflicts / pdm_div) + 1) == 0) ||
			(!pdm_freq && SH < 2 && starts % maxConflicts == 0) ||
			(pdm_freq && starts % pdm_freq == 0)) {
			ref_vars = UNKNOWN;
			R = pdm_rounds;
			//printf("pdm_div = %d, maxConflicts = %d, starts = %d\n", pdm_div, maxConflicts, starts);
		}
		if (maxConflicts > 50000) pdm_div += 5;
	}
	inline void clHist(const G_REF& gref) {
		B_REF c = (B_REF)gref;
		for (LIT_POS i = 0; i < c->size(); i++) {
			assert((*c)[i] > 0);
			if (ISNEG((*c)[i])) occurs[V2X((*c)[i])].ns++;
			else occurs[V2X((*c)[i])].ps++;
		}
	}
	inline void clBumpAct(C_REF& c) {
		c->inc_act(cl_params.cl_inc);
		if (c->activity() > (float)1e30) {
			for (int i = 0; i < learnts.size(); i++) learnts[i]->sca_act((float)1e-30);
			cl_params.cl_inc *= (float)1e-30;
		}
	}
	inline double luby_seq(double y, int x) {
		int size, seq;
		for (size = 1, seq = 0; size < x + 1; seq++, size = (size << 1) + 1);
		while (size - 1 != x) {
			size = (size - 1) >> 1;
			seq--;
			x = x % size;
		}
		return pow(y, seq);
	}
	inline void printWatched(const uint32& v_idx) {
		printf("c |\t\tWL of v(%d)\n", v_idx + 1);
		uint32 p = V2D(v_idx + 1);
		wt.print(p), wt.print(NEG(p));
		printf("c |---------------------------------\n");
	}
	inline void printStats() {
		if (verbose == 1) {
			int remainedVars = (int)nOrgVars() - (int)nRemVars();
			printf("c | %9d %9d %10lld | %10lld %10lld %10lld %10lld %7.0f |\n",
				remainedVars, nClauses(), nLiterals(),
				nConflicts, lrn.max_learnt_cls,
				(int64)learnts.size(), nLearntLits(), (float)nLearntLits() / learnts.size());
		}
	}
	inline void printStats(bool p) {
		if (verbose == 1 && p) {
			int remainedVars = (int)nOrgVars() - (int)nRemVars();
			int l2c = learnts.size() == 0 ? 0 : int(nLearntLits() / learnts.size());
			printf("c | %9d %9d %9d %10lld %6d %9d %8d %10lld %6d |\n",
				remainedVars, nClauses(), nBins(), nLiterals(),
				starts,
				learnts.size(), nGlues(), nLearntLits(), l2c);
			progRate += (progRate >> 2);
		}
	}
	inline void write_proof(uint32* lits, const CL_LEN& len) {
		assert(len > 0);
		uint32* lit = lits, * end = lits + len;
		while (lit != end) {
			register uint32 b = 0;
			if (mapped) b = revLit(*lit++);
			else b = *lit++;
			while (b > 127) { write_proof(Byte(128 | (b & 127))); b >>= 7; }
			write_proof(Byte(b));
		}
	}
	//=============================================
	// helpers
	inline void printLit(const uint32& l) { printf("%d@%d", (ISNEG(l)) ? -int(ABS(l)) : ABS(l), sol->level(ABS(l) - 1)); }
	inline void printVars(const uint32* arr, const int& size)
	{
		printf("c | [");
		for (int i = 0; i < size; i++) {
			printf("%4d ", arr[i] + 1);
		}
		printf("]\n");
	}
	inline void printAssigns(const uint32* arr, const int& size)
	{
		printf("c | [");
		for (int i = 0; i < size; i++) {
			printf("%4d ", ISNEG(arr[i]) ? -int(ABS(arr[i])) : ABS(arr[i]));
		}
		printf("]\n");
	}
	inline void printAssigns(void)
	{
		for (int i = 0; i < sp->assigns.size(); i++)
			printf("c | trail[%d] = %d@%d\n", i, ISNEG(sp->assigns[i]) ? -int(ABS(sp->assigns[i])) : ABS(sp->assigns[i]), sol->level(V2X(sp->assigns[i])));
	}
	inline void printLearnt(void)
	{
		printf("c | Learnt(");
		for (LIT_POS n = 0; n < learnt_cl.size(); n++) {
			printf(" %d@%d ", (ISNEG(learnt_cl[n])) ? -int(ABS(learnt_cl[n])) : ABS(learnt_cl[n]), sol->level(V2X(learnt_cl[n])));
		}
		printf(")\n");
	}
	inline void printClause(const uVec1D& cl)
	{
		printf("c | (");
		for (int i = 0; i < cl.size(); i++) {
			printf("%4d ", ISNEG(cl[i]) ? -int(ABS(cl[i])) : ABS(cl[i]));
		}
		printf(")\n");
	}
	template<class T>
	inline void printCNF(const Vec<T>& cnf, const int& offset = 0) {
		printf("c |               Host CNF(sz = %d)\n", cnf.size());
		for (int c = offset; c < cnf.size(); c++) {
			if (cnf[c]->size() > 0) {
				printf("c | C(%d)->", c);
				cnf[c]->print();
			}
		}
		printf("c |=================================================|\n");
	}
	inline void printOL(const OL& list) {
		for (int i = 0; i < list.size(); i++) {
			printf("c | ");
			list[i]->print();
		}
	}
	inline void printOT(void) {
		printf("c | Positive occurs:\n");
		for (uint32 v = 0; v < nOrgVars(); v++) {
			uint32 p = V2D(v + 1);
			if (ot[p].size() != 0) {
				printf("c | Var(%d):\n", v + 1);
				printOL(ot[p]);
			}
		}
		printf("c | Negative occurs:\n");
		for (uint32 v = 0; v < nOrgVars(); v++) {
			uint32 n = NEG(V2D(v + 1));
			if (ot[n].size() != 0) {
				printf("c | Var(%d):\n", -int(v + 1));
				printOL(ot[n]);
			}
		}
	}
	//==============================================
	int64 sysMemUsed(void);
	void sysFree(void);
	void killSolver(const CNF_STATE& status = TERMINATE);
	void resetSolver(const bool& re = true);
	void allocSolver(const bool& re = false);
	void initSolver(void);
	void recycle(void);
	void cnfrewriter(const string&);
	void varOrder(void);
	void hist(BCNF&, const bool& rst = false);
	void hist(LCNF&, const bool& rst = false);
	void hist(WT&, const bool& rst = false);
	void attachClause(B_REF&);
	void attachClause(C_REF&);
	void shrinkClause(G_REF);
	void removeClause(B_REF&, const bool& gc = true);
	void removeClause(C_REF&, const bool& gc = true);
	void detachClause(B_REF&, const bool& gc = true);
	int simplify(BCNF&);
	int simplify(LCNF&);
	void simplify(void);
	void simplify_top(void);
	void solve(void);
	CNF_STATE parser(const string&);
	CNF_STATE search(void);
	CNF_STATE decide(void);
	void pumpFrozen(void);
	void PDMInit(void);
	void PDM(void);
	bool depFreeze_init(WL&, const uint32&);
	bool depFreeze(WL&, const uint32&);
	void eligibility(void);
	G_REF BCP(void);
	void analyze(G_REF&);
	void cbtLevel(G_REF&);
	void btLevel(void);
	void binSelfsub(void);
	void selfsub(void);
	bool selfsub(const uint32&, uint32*, CL_LEN&, const uint32&);
	void cancelAssigns(const int&);
	void backJump(const int&);
	void lReduce(void);
	bool consistent(BCNF&, WT&);
	bool consistent(LCNF&, WT&);
	void printReport(void);
	void printModel(void);
	void wrapUp(const CNF_STATE&);
	/* flags */
	bool quiet_en, parse_only_en, rewriter_en, perf_en;
	bool mcv_en, model_en, proof_en, fdp_en, cbt_en;
	int progRate,initProgRate, verbose, seed, timeout;
	int pdm_rounds, pdm_freq, pdm_order;
	int SH, polarity, restart_base, blRestMin, lbdRestBase, blockRestBase, VSIDSDecayFreq;
	int nClsReduce, lbdFrozen, lbdMinReduce, lbdMinClSize, incReduceSmall, incReduceBig;
	int cbt_dist, cbt_conf_max;
	double var_inc, var_decay;
	double RF, RB, restart_inc, gperc;
	string restPolicy, proof_path;
	/********************************************/
	/*                Simplifier                */
	/********************************************/
protected:
	uVec1D removed; Vec1D mappedVars, reverseVars;
	SCNF scnf; OT ot;
	bool mapped;
public:
	//============== inline methods ===============
	inline void collectClause(S_REF& c) { assert(c != NULL);  delete c; c = NULL; }
	inline uint32 mapLit(const uint32& orgLit) {
		static int maxVar = 1;
		assert(!mappedVars.empty());
		register int orgVar = ABS(orgLit);
		assert(!pv->melted[orgVar - 1]);
		assert(orgVar > 0 && orgVar < mappedVars.size());
		if (mappedVars[orgVar] == 0) { reverseVars[maxVar] = orgVar; mappedVars[orgVar] = maxVar++; }
		assert(maxVar - 1 <= (int)nOrgVars());
		return (V2D(mappedVars[orgVar]) | ISNEG(orgLit));
	}
	inline uint32 revLit(const uint32& mapLit) { return (V2D(reverseVars[ABS(mapLit)]) | ISNEG(mapLit)); }
	inline void mapClause(SCLAUSE& s, BCLAUSE& d) {
		assert(d.size());
		for (LIT_POS l = 0; l < s.size(); l++) d[l] = mapLit(s[l]);
	}
	inline void clHist(const S_REF& c) {
		assert(c->size());
		assert(c->status() == ORIGINAL || c->status() == LEARNT);
		for (LIT_POS i = 0; i < c->size(); i++) {
			assert(c->lit(i));
			if (ISNEG(c->lit(i))) occurs[V2X(c->lit(i))].ns++;
			else occurs[V2X(c->lit(i))].ps++;
		}
	}
	inline void countCls(void) {
		cnf_stats.n_cls_after = 0;
		cnf_stats.n_lits_after = 0;
		for (int i = 0; i < scnf.size(); i++)
			if (scnf[i]->status() != DELETED) {
				cnf_stats.n_cls_after++;
				cnf_stats.n_lits_after += scnf[i]->size();
			}
	}
	inline void countLits(void) {
		cnf_stats.n_lits_after = 0;
		for (int i = 0; i < scnf.size(); i++)
			if (scnf[i]->status() != DELETED)
				cnf_stats.n_lits_after += scnf[i]->size();
	}
	inline void evalReds(void) {
		cnf_stats.n_del_vars = 0;
		cnf_stats.n_cls_after = 0;
		cnf_stats.n_lits_after = 0;
		for (int i = 0; i < scnf.size(); i++) {
			if (scnf[i]->status() != DELETED) {
				cnf_stats.n_cls_after++;
				cnf_stats.n_lits_after += scnf[i]->size();
				for (LIT_POS k = 0; k < scnf[i]->size(); k++)
					sp->frozen[V2X(scnf[i]->lit(k))] = true;
			}
		}
		bool* f_end = sp->frozen + nOrgVars();
		for (bool* f = sp->frozen; f != f_end; f++) {
			if (*f == 0) cnf_stats.n_del_vars++;
			else *f = 0;
		}
	}
	inline void logReductions(void) {
		printf("c |  Parallel variables = %d\n", pv->nPVs);
		printf("c |  Variables after = %d (-%d)\n", nOrgVars() - cnf_stats.n_del_vars, cnf_stats.n_del_vars);
		printf("c |  Clauses after = %d (-%d)\n", cnf_stats.n_cls_after, nClauses() - cnf_stats.n_cls_after);
		if (cnf_stats.n_lits_after > nLiterals()) printf("c |  Literals after = %lld (+%lld)\n", cnf_stats.n_lits_after, cnf_stats.n_lits_after - nLiterals());
		else printf("c |  Literals after = %lld (-%lld)\n", cnf_stats.n_lits_after, nLiterals() - cnf_stats.n_lits_after);
	}
	inline void printPStats() {
		if (verbose == 1 && SH == 2) {
			int remainedVars = (int)nOrgVars() - (int)nRemVars();
			printf("c |p %8d %9d %9s %10lld %6d %9s %8s %10s %6s |\n",
				remainedVars, nClauses(), "---", nLiterals(), starts, "---", "---", "---", "---");
		}
		if (verbose == 1 && SH < 2) {
			int remainedVars = (int)nOrgVars() - (int)nRemVars();
			printf("c |p %8d %9d %10lld | %10lld %10s %10s %10s %7s |\n",
				remainedVars, nClauses(), nLiterals(), nConflicts, "---", "---", "---", "---");
		}
	}
	//==============================================
	void optSimp();
	void hist(const SCNF&, const bool& rst = false);
	void varReorder(void);
	void extractBins(void);
	bool awaken(void);
	void cleanSlate(void);
	void reduce_ol(OL&);
	void reduce_ot(void);
	void create_ot(const bool& rst = true);
	CNF_STATE prop(void);
	void strengthen(S_REF, const uint32&);
	bool propClause(S_REF, const uint32&);
	void attachClause(S_REF, const bool& added = true);
	void reattachClause(B_REF);
	void shrinkCNF(const bool& countVars = false);
	void _simplify(void);
	void preprocess(void);
	void depFreeze(const OL&, const uint32&, const int&, const int&);
	void LCVE(void);
	void bve(void);
	void VE(void);
	void HSE(void);
	void HRE(void);
	void BCE(void);
	void SUB(void);
	bool consistent(const SCNF&, const OT&);
	/* flags */
	bool pre_en, lpre_en, ve_en, ve_plus_en, sub_en, bce_en, hre_en, all_en;
	int pre_delay, mu_pos, mu_neg, phases, cnf_free_freq;
};
extern ParaFROST* g_pFrost;

#endif 